package RT::Extension::SaltedPasswords;

use strict;
use warnings;

our $VERSION = "1.2";
use Encode;
use Digest::MD5;

sub IsPassword {
    my $self  = shift;
    my $value = shift;

    # RT does not allow null passwords
    if ( ( !defined($value) ) or ( $value eq '' ) ) {
        return (undef);
    }

   if ( $self->PrincipalObj->Disabled ) {
        $RT::Logger->info(
            "Disabled user " . $self->Name . " tried to log in" );
        return (undef);
    }

    unless ($self->HasPassword) {
        return(undef);
     }

    my $stored = $self->__Value('Password');
    if (length $stored == 40) {
        # The truncated SHA256(salt,MD5(passwd)) form from 2010/12 is 40 characters long
        my $hash = MIME::Base64::decode_base64($stored);
        # The first 4 bytes are the salt, the rest is substr(SHA256,0,26)
        my $salt = substr($hash, 0, 4, "");
        return substr(sha256($salt . Digest::MD5::md5($value)), 0, 26) eq $hash;
    } elsif (length $stored == 32) {
        # Hex nonsalted-md5
        return 0 unless Digest::MD5::md5_hex(Encode::encode_utf8($value)) eq $stored;
    } elsif (length $stored == 22) {
        # Base64 nonsalted-md5
        return 0 unless Digest::MD5::md5_base64(Encode::encode_utf8($value)) eq $stored;
    } elsif (length $stored == 13) {
        # crypt() output
        return 0 unless crypt(Encode::encode_utf8($value), $stored) eq $stored;
    } else {
        $RT::Logger->warn("Unknown password form");
        return 0;
    }

    # We got here by validating successfully, but with a legacy
    # password form.  Update to the most recent form.
    my $obj = $self->isa("RT::CurrentUser") ? $self->UserObj : $self;
    $obj->_Set(Field => 'Password', Value =>  $self->_GeneratePassword($value) );
    return 1;
}

sub _GeneratePassword {
    my $self = shift;
    my ($password, $salt) = @_;

    # Generate a random 4-byte salt
    $salt ||= pack("C4",map{int rand(256)} 1..4);

    # Encode the salt, and a truncated SHA256 of the MD5 of the
    # password.  The additional, un-necessary level of MD5 allows for
    # transparent upgrading to this scheme, from the previous unsalted
    # MD5 one.
    return MIME::Base64::encode_base64(
        $salt . substr(sha256($salt . Digest::MD5::md5($password)),0,26),
        "" # No newline
    );
}

BEGIN {
    if ($RT::VERSION =~ /^3\.(6\.\d+|8\.[0-8])$/) {
        no warnings 'redefine';
        *RT::User::_GeneratePassword = \&_GeneratePassword;
        *RT::User::IsPassword = \&IsPassword;

        # Determine which SHA256 implementation to use
        if ( eval { require Digest::SHA; 1} and not $@) {
            # The C implementation is much preferred
            *sha256 = \&Digest::SHA::sha256;
        } else {
            require RT::Digest::SHA::PurePerl;
            if ( eval { require Digest::SHA::PurePerl; 1} and not $@
                 and $Digest::SHA::PurePerl::VERSION >= $RT::Digest::SHA::PurePerl::VERSION) {
                # Prefer the non-bundled version, particularly if it's more recent
                *sha256 = \&Digest::SHA::PurePerl::sha256;
            } else {
                *sha256 = \&RT::Digest::SHA::PurePerl::sha256;
            }
        }
    }
}

__END__

=head1 NAME

RT::Extension::SaltedPasswords - Security hotpatch for RT < 3.8.9

=head1 AUTHOR

Alex Vandiver C<< <alexmv@bestpractical.com> >>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2010, Best Practical Solutions, LLC.  All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.

=cut

1;
